/*
 * ============================================================================
 *  SYM_TBL:        Assembler Symbol Table Module
 *
 *  Author:         J. Zbiciak
 *
 *  $Id: sym_tbl.c,v 1.5 1999/09/12 10:28:03 im14u2c Exp $
 * ============================================================================
 *  The symbol table management routines, as well as the initial global
 *  symbol table are contained in this file.
 * ============================================================================
 *  SYM_INIT     -- Initialize global symbol table
 *  SYM_SETSYM   -- Sets a symbol's flags, value.
 *  SYM_GETALL   -- Returns entire encoded flag-set for a symbol.
 *  SYM_GETTYPE  -- Get type of symbol (instruction, directive, label, etc.)
 *  SYM_GETOP    -- Get opcode encoding for symbol (if any)
 *  SYM_GETDIR   -- Get directive number corresponding to symbol (if any)
 *  SYM_GETSTX   -- Get syntax for instruction (if any)
 *  SYM_GETVAL   -- Returns the value associated w/ a symbol.
 *  SYM_GETREFS  -- Return list of line numbers which refer to symbol.
 *  SYM_EQUATE   -- Equate two symbols.
 *  SYM_RESOLVE  -- Perform final symbol-resolution pass over symbol table.
 *  SYM_UNRSLVD  -- Return a sorted list of unresolved symbols.
 *  SYM_HASH     -- Symbol hashing function.  0 <= hash < SYM_HASH_SZ
 * ============================================================================
 */

#include "../config.h"
#include "asm.h"
#include "sym_tbl.h"

static const char rcs_id[] = "$Id: sym_tbl.c,v 1.5 1999/09/12 10:28:03 im14u2c Exp $";

/*
 * ============================================================================
 *  SYM_TBL      -- Global Symbol Table
 * ============================================================================
 */

sym_tbl_t sym_tbl = { 0, {0}, {0} };

/*
 * ============================================================================
 *  SYM_TBL_DFL  -- Default symbol table contents
 *
 *  This symbol table contains all of the defaults for the assembler, 
 *  all of the assembly mnemonics, pseudo-ops, directives, and built-in
 *  symbols.
 *
 *  Note that this table is static and is inserted into the global symbol
 *  table at initialization time.  Also, "FLAG_FIXED" is set at init
 *  time rather than here, as a notational convenience.
 * ============================================================================
 */

static sym_rec_t sym_tbl_dfl[] =
{
    /* -------------------------------------------------------------------- */
    /*  Handy table-building macros...                                      */
    /*  REC  --  saves some typing by filling in the unused fields w NULLs  */
    /*  DIR  --  Fills in the directive field in the flags.                 */
    /*  OP   --  Fills in the opcode type field in the flags.               */
    /* -------------------------------------------------------------------- */
    #define REC(s,v,f)  { NULL, s, NULL, NULL, v, f, {0}} 
    #define DIR(d)      SYM_SETFLD(FLAG_DIR,  SYM_FLD_DIR, SYM_SHF_DIR, d)
    #define OP(o,s)     SYM_SETFLD(\
                        SYM_SETFLD(FLAG_INSN, SYM_FLD_OP,  SYM_SHF_OP , o), \
                                              SYM_FLD_STX, SYM_SHF_STX, s)

    /* -------------------------------------------------------------------- */
    /*  Directives.                                                         */
    /* -------------------------------------------------------------------- */
    REC("ORG",      0x0000,         DIR(DIR_ORG)),
    REC("EQU",      0x0000,         DIR(DIR_EQU)),
    REC("DCW",      0x0000,         DIR(DIR_DCW)),

    /* -------------------------------------------------------------------- */
    /*  Instruction Mnemonics, with opcode and syntax tables.               */
    /*                                                                      */
    /*  Note, for most of these instructions, the opcode template is just   */
    /*  kept in bits 9:0.  For Jumps, the fields "bb" and "ii" are kept     */
    /*  in fields 11:10 and 13:12, respectively.  These are encoded in the  */
    /*  second decle of the instruction when actually emitted to the        */
    /*  object file.                                                        */
    /*                                                                      */
    /*  See the "cp_1610.txt" documentation for opcodes.                    */
    /* -------------------------------------------------------------------- */
    /* Implied 1-Op */
    REC("HLT",      0x0000,         OP(OP_IMPL_1OP,     STX_IMPL    )),
    REC("SDBD",     0x0001,         OP(OP_IMPL_1OP,     STX_IMPL    )),
    REC("EIS",      0x0002,         OP(OP_IMPL_1OP,     STX_IMPL    )),
    REC("DIS",      0x0003,         OP(OP_IMPL_1OP,     STX_IMPL    )),
    REC("TCI",      0x0005,         OP(OP_IMPL_1OP,     STX_IMPL    )),
    REC("CLRC",     0x0006,         OP(OP_IMPL_1OP,     STX_IMPL    )),
    REC("SETC",     0x0007,         OP(OP_IMPL_1OP,     STX_IMPL    )),

    /* Jump, Jump-to-SubRoutine */
    REC("J",        0x0C04,         OP(OP_JUMP,         STX_JUMP    )),
    REC("JSR",      0x0004,         OP(OP_JUMP,         STX_JSR     )),
    REC("JE",       0x1C04,         OP(OP_JUMP,         STX_JUMP    )),
    REC("JSRE",     0x1004,         OP(OP_JUMP,         STX_JSR     )),
    REC("JD",       0x2C04,         OP(OP_JUMP,         STX_JUMP    )),
    REC("JSRD",     0x2004,         OP(OP_JUMP,         STX_JSR     )),

    /* Register 1-OP */
    REC("INCR",     0x0008,         OP(OP_REG_1OP,      STX_1OP_R   )),
    REC("DECR",     0x0010,         OP(OP_REG_1OP,      STX_1OP_R   )),
    REC("COMR",     0x0018,         OP(OP_REG_1OP,      STX_1OP_R   )),
    REC("NEGR",     0x0020,         OP(OP_REG_1OP,      STX_1OP_R   )),
    REC("ADCR",     0x0028,         OP(OP_REG_1OP,      STX_1OP_R   )),
    REC("GSWD",     0x0030,         OP(OP_GSWD,         STX_1OP_R   )),
    REC("RSWD",     0x0038,         OP(OP_REG_1OP,      STX_1OP_R   )),

    /* NOP, SIN */
    REC("NOP",      0x0034,         OP(OP_NOP_SIN,      STX_1OP_12  )),
    REC("SIN",      0x0034,         OP(OP_NOP_SIN,      STX_1OP_12  )),

    /* Rotates and Shifts */
    REC("SWAP",     0x0040,         OP(OP_ROT_1OP,      STX_1OP_R12 )),
    REC("SLL",      0x0048,         OP(OP_ROT_1OP,      STX_1OP_R12 )),
    REC("RLC",      0x0050,         OP(OP_ROT_1OP,      STX_1OP_R12 )),
    REC("SLLC",     0x0058,         OP(OP_ROT_1OP,      STX_1OP_R12 )),
    REC("SLR",      0x0060,         OP(OP_ROT_1OP,      STX_1OP_R12 )),
    REC("SAR",      0x0068,         OP(OP_ROT_1OP,      STX_1OP_R12 )),
    REC("RRC",      0x0070,         OP(OP_ROT_1OP,      STX_1OP_R12 )),
    REC("SARC",     0x0078,         OP(OP_ROT_1OP,      STX_1OP_R12 )),

    /* Register/Register 2-Op */
    REC("MOVR",     0x0180,         OP(OP_REG_2OP,      STX_2OP_RR  )),
    REC("ADDR",     0x01C0,         OP(OP_REG_2OP,      STX_2OP_RR  )),
    REC("SUBR",     0x0200,         OP(OP_REG_2OP,      STX_2OP_RR  )),
    REC("CMPR",     0x0240,         OP(OP_REG_2OP,      STX_2OP_RR  )),
    REC("ANDR",     0x0280,         OP(OP_REG_2OP,      STX_2OP_RR  )),
    REC("XORR",     0x02C0,         OP(OP_REG_2OP,      STX_2OP_RR  )),

    /* Conditional Branches */
    REC("B",        0x0200,         OP(OP_COND_BR,      STX_COND_BR )),
    REC("BC",       0x0201,         OP(OP_COND_BR,      STX_COND_BR )),
    REC("BOV",      0x0202,         OP(OP_COND_BR,      STX_COND_BR )),
    REC("BPL",      0x0203,         OP(OP_COND_BR,      STX_COND_BR )),
    REC("BZE",      0x0204,         OP(OP_COND_BR,      STX_COND_BR )),
    REC("BLT",      0x0205,         OP(OP_COND_BR,      STX_COND_BR )),
    REC("BLE",      0x0206,         OP(OP_COND_BR,      STX_COND_BR )),
    REC("BUSC",     0x0207,         OP(OP_COND_BR,      STX_COND_BR )),
    REC("NOPP",     0x0208,         OP(OP_COND_BR,      STX_IMPL    )),
    REC("BNC",      0x0209,         OP(OP_COND_BR,      STX_COND_BR )),
    REC("BNOV",     0x020A,         OP(OP_COND_BR,      STX_COND_BR )),
    REC("BMI",      0x020B,         OP(OP_COND_BR,      STX_COND_BR )),
    REC("BNZE",     0x020C,         OP(OP_COND_BR,      STX_COND_BR )),
    REC("BGE",      0x020D,         OP(OP_COND_BR,      STX_COND_BR )),
    REC("BGT",      0x020E,         OP(OP_COND_BR,      STX_COND_BR )),
    REC("BESC",     0x020F,         OP(OP_COND_BR,      STX_COND_BR )),

    /* Conditional Branch Aliases */
    REC("BEQ",      0x0204,         OP(OP_COND_BR,      STX_COND_BR )),
    REC("BNGE",     0x0205,         OP(OP_COND_BR,      STX_COND_BR )),
    REC("BNGT",     0x0206,         OP(OP_COND_BR,      STX_COND_BR )),
    REC("BNEQ",     0x020C,         OP(OP_COND_BR,      STX_COND_BR )),
    REC("BNLT",     0x020D,         OP(OP_COND_BR,      STX_COND_BR )),
    REC("BNLE",     0x020E,         OP(OP_COND_BR,      STX_COND_BR )),

    /* Direct/Register 2-Op */
    REC("MVO",      0x0240,         OP(OP_DIR_2OP,      STX_2OP_RL  )),
    REC("MVI",      0x0280,         OP(OP_DIR_2OP,      STX_2OP_LR  )),
    REC("ADD",      0x02C0,         OP(OP_DIR_2OP,      STX_2OP_LR  )),
    REC("SUB",      0x0300,         OP(OP_DIR_2OP,      STX_2OP_LR  )),
    REC("CMP",      0x0340,         OP(OP_DIR_2OP,      STX_2OP_LR  )),
    REC("AND",      0x0380,         OP(OP_DIR_2OP,      STX_2OP_LR  )),
    REC("XOR",      0x03C0,         OP(OP_DIR_2OP,      STX_2OP_LR  )),

    /* Indirect/Register 2-Op */
    REC("MVO@",     0x0240,         OP(OP_IND_2OP,      STX_2OP_RR  )),
    REC("MVI@",     0x0280,         OP(OP_IND_2OP,      STX_2OP_RR  )),
    REC("ADD@",     0x02C0,         OP(OP_IND_2OP,      STX_2OP_RR  )),
    REC("SUB@",     0x0300,         OP(OP_IND_2OP,      STX_2OP_RR  )),
    REC("CMP@",     0x0340,         OP(OP_IND_2OP,      STX_2OP_RR  )),
    REC("AND@",     0x0380,         OP(OP_IND_2OP,      STX_2OP_RR  )),
    REC("XOR@",     0x03C0,         OP(OP_IND_2OP,      STX_2OP_RR  )),

    /* Immediate/Register 2-Op */
    REC("MVOI",     0x0240,         OP(OP_IMM_2OP,      STX_2OP_RK  )),
    REC("MVII",     0x0280,         OP(OP_IMM_2OP,      STX_2OP_KR  )),
    REC("ADDI",     0x02C0,         OP(OP_IMM_2OP,      STX_2OP_KR  )),
    REC("SUBI",     0x0300,         OP(OP_IMM_2OP,      STX_2OP_KR  )),
    REC("CMPI",     0x0340,         OP(OP_IMM_2OP,      STX_2OP_KR  )),
    REC("ANDI",     0x0380,         OP(OP_IMM_2OP,      STX_2OP_KR  )),
    REC("XORI",     0x03C0,         OP(OP_IMM_2OP,      STX_2OP_KR  )),

    /* -------------------------------------------------------------------- */
    /*  Pseudo-Ops                                                          */
    /*                                                                      */
    /*      TSTR Rx  -->  MOVR Rx, Rx                                       */
    /*      JR Rx    -->  MOVR Rx, R7                                       */
    /*      CLRR Rx  -->  XORR Rx, Rx                                       */
    /*      PSHR Rx  -->  MVO@ Rx, R6                                       */
    /*      PULR Rx  -->  MVI@ R6, Rx                                       */
    /* -------------------------------------------------------------------- */
    REC("TSTR",     0x0180,         OP(OP_RPT_2OP,      STX_1OP_R   )),
    REC("CLRR",     0x02C0,         OP(OP_RPT_2OP,      STX_1OP_R   )),
    REC("PSHR",     0x0246,         OP(OP_REG_2OP_DI,   STX_1OP_R   )),
    REC("PULR",     0x02B0,         OP(OP_REG_1OP,      STX_1OP_R   )),
    REC("JR",       0x0247,         OP(OP_REG_2OP_DI,   STX_1OP_R   )),

    /* -------------------------------------------------------------------- */
    /*  Registers                                                           */
    /* -------------------------------------------------------------------- */
    REC("R0",       0x0000,         FLAG_REG),
    REC("R1",       0x0001,         FLAG_REG),
    REC("R2",       0x0002,         FLAG_REG),
    REC("R3",       0x0003,         FLAG_REG),
    REC("R4",       0x0004,         FLAG_REG),
    REC("R5",       0x0005,         FLAG_REG),
    REC("R6",       0x0006,         FLAG_REG),
    REC("R7",       0x0007,         FLAG_REG),
    REC("SP",       0x0006,         FLAG_REG),
    REC("PC",       0x0007,         FLAG_REG),
};


/*
 * ============================================================================
 *  SYM_HASH     -- Symbol hashing function.  0 <= hash < SYM_HASH_SZ
 * ============================================================================
 */
uint_32 sym_hash(const char *name)
{
    uint_32 hash_val = 0x2A3A4A5A, c;
    int len = 0;

    while ((c = *name++) != '\0')
    {
        len++;
        hash_val = (hash_val << 3) - (hash_val >> 7) + c;
    }

    hash_val += (len + 3) * 0x2A3A4A5A;

    return hash_val % SYM_HASH_SZ;
}

/*
 * ============================================================================
 *  SYM_SETSYM   -- Sets a symbol's flags, value.
 *
 *  This is the Swiss-Army-Knife of symbol-table utilities.
 *
 *   -- All label refs and defs call this with flag == FLAG_LABEL.
 *   -- References val == 0, and def == 0.
 *   -- Definitions call this with val == (label's value), and def == 1.
 *
 *  All label defs and refs call this so that unresolved symbols are
 *  recorded and symbols are resolved where possible.  At the end of 
 *  "pass 1", all symbols should be resolvable.
 * ============================================================================
 */

int                     /* 0 == success or -1 == failure        */
sym_setsym

    char    *name,      /* Symbol name                          */
    uint_32 flag,       /* Flags associated w/ symbol           */
    uint_16 val,        /* Value to set symbol to.              */
    int     def,        /* Whether this is a def or ref.        */
    char    *file,      /* File name of ref or def (if any).    */
    uint_32 line_no     /* Line number of ref or def.           */
)
{
    int         i;
    uint_32     hash, ehash;
    sym_rec_p   rec = NULL, *prnt, equ_rec;

    /* -------------------------------------------------------------------- */
    /*  Generate a hash-value for the symbol.                               */
    /* -------------------------------------------------------------------- */
    hash = sym_hash(name);

    /* -------------------------------------------------------------------- */
    /*  Probe the symbol table and see if we can find the symbol.           */
    /* -------------------------------------------------------------------- */
    prnt = &sym_tbl.hash[hash]; 
    rec  =  sym_tbl.hash[hash];
    while (rec != NULL && strcasecmp(rec->symb, name))
    {
        prnt = &rec->link; 
        rec  =  rec->link;
    }

    /* -------------------------------------------------------------------- */
    /*  If we didn't find a symbol, allocate a hash chain link and append   */
    /*  it to the hash-chain.  Increment the chain length for that chain    */
    /*  also.                                                               */
    /* -------------------------------------------------------------------- */
    if (!rec)
    {
        if (!(rec = (sym_rec_p)calloc(1, sizeof(sym_rec_t))))
            die("Out of memory in sym_tbl:sym_setsym()\n");

        /* ---------------------------------------------------------------- */
        /*  If this is not a 'def', force the flags to contain the          */
        /*  unresolved bit, since 'ref's cannot resolve a symbol.           */
        /* ---------------------------------------------------------------- */
        if (!def) flag |= FLAG_UNRSLVD;

        rec->name = strdup(name);
        rec->flag = flag;
        rec->val  = val;
        *prnt     = rec;

        sym_tbl.deep[hash]++;
    }

    /* -------------------------------------------------------------------- */
    /*  Add the line-number reference to the list of references.            */
    /* -------------------------------------------------------------------- */
    if (rec->num_ref >= MAXREFS)
    {
        char buf[1024];
        sprintf(buf, "Too many references to symbol '%s'.\n", name);
        die(buf);
    }

    /* -------------------------------------------------------------------- */
    /*  Merge the flags with the existing flags.                            */
    /* -------------------------------------------------------------------- */
    rec->flag |= flag;




    /* NOT DONE */
}


/* ======================================================================== */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       */
/*  General Public License for more details.                                */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               */
/* ======================================================================== */
/*                 Copyright (c) 1998-1999, Joseph Zbiciak                  */
/* ======================================================================== */

