/*
 * ============================================================================
 *  Title:    Platform Portability "Library"
 *  Author:   J. Zbiciak
 *  $Id: plat_lib.h,v 1.2 1999/09/14 06:29:30 tlindner Exp $
 * ============================================================================
 *  This module fills in missing features on various platforms.
 * ============================================================================
 *  GETTIMEOFDAY     -- Return current time in seconds/microseconds.
 *  STRDUP           -- Copy a string into freshly malloc'd storage.
 *  SNPRINTF         -- Like sprintf(), only with bounds checking.
 *  PLAT_DELAY       -- Sleep w/ millisecond precision.
 * ============================================================================
 */

#ifndef _PLAT_LIB_H
#define _PLAT_LIB_H
 
/* ------------------------------------------------------------------------ */
/*  GETTIMEOFDAY     -- Return current time in seconds/microseconds.        */
/*                                                                          */
/*  Cheesy gettimeofday() implementation for platforms that lack one.       */
/*  If 'clock()' is available, we use that in a macro to implement the      */
/*  gettimeofday() function.  Otherwise, we fall back on a function call,   */
/*  in the hopes that we have a platform-specific version in "port_lib.c".  */
/* ------------------------------------------------------------------------ */
#ifdef NO_GETTIMEOFDAY

struct timeval 
{
    long tv_sec;        /* seconds */
    long tv_usec;  /* microseconds */
};

#if defined(CLK_TCK) && !defined(NO_CLOCK)
# define TV_CLK_TCK (int)CLK_TCK
# define gettimeofday(x,y) do {                                             \
        struct timeval *___tv = (x);                                        \
        long            ___now = clock();                                   \
                                                                            \
        ___tv->tv_sec  = ___now / TV_CLK_TCK;                               \
        ___tv->tv_usec = (___now % TV_CLK_TCK) * (1000000 / TV_CLK_TCK);    \
    } while (0)
#else
void gettimeofday(struct timeval *, int *);
#endif

#endif /* NO_GETTIMEOFDAY */


/* ------------------------------------------------------------------------ */
/*  STRDUP           -- Copy a string into freshly malloc'd storage.        */
/*                                                                          */
/*  Unfortunately, strdup() is not specified by ANSI.  *sigh*               */
/* ------------------------------------------------------------------------ */
#ifdef NO_STRDUP
char * strdup(const char *s);
#endif /* NO_STRDUP */

/* ------------------------------------------------------------------------ */
/*  SNPRINTF         -- Like sprintf(), only with bounds checking.          */
/* ------------------------------------------------------------------------ */
/*  WARNING:  THIS COULD CAUSE BUFFER OVERFLOW PROBLEMS AND IS MERELY       */
/*            A SHIM WHICH IS BEING USED TO GET jzIntv TO COMPILE.          */
/* ------------------------------------------------------------------------ */
#ifdef NO_SNPRINTF
void snprintf(char * buf, int len, const char * fmt, ...);
#endif /* NO_SNPRINTF */

/* ------------------------------------------------------------------------ */
/*  PLAT_DELAY       -- Sleep w/ millisecond precision.                     */
/* ------------------------------------------------------------------------ */
#ifdef NO_SDL
void plat_delay(unsigned delay);
#else
# define plat_delay SDL_Delay
#endif

#endif /* PORT_LIB_H */

/* ======================================================================== */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       */
/*  General Public License for more details.                                */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               */
/* ======================================================================== */
/*                 Copyright (c) 1998-1999, Joseph Zbiciak                  */
/* ======================================================================== */
